import abc
import datetime as dt

import pandas as pd

from analytix import InvalidRequest, NoAuthorisedService


class YouTubeAnalyticsReport:
    __slots__ = ("data", "ncolumns", "nrows")

    def __init__(self, data):
        self.data = data
        self.ncolumns = len(data["columnHeaders"])
        self.nrows = len(data["rows"])

    def to_dataframe(self):
        df = pd.DataFrame()
        df = df.append(self.data["rows"])
        df.columns = [c["name"] for c in self.data["columnHeaders"]]
        return df

    def to_json(self, file, indent=4):
        if not file.endswith(".json"):
            file += ".json"
        with open(file, "w", encoding="utf-8") as f:
            json.dump(self.data, f, indent=indent, ensure_ascii=False)

    def to_csv(self, file):
        if not file.endswith(".csv"):
            file += ".csv"
        self.to_dataframe().to_csv(file)


class YouTubeAnalytics(metaclass=abc.ABCMeta):
    __slots__ = ("service", "metrics", "dimensions", "filters")

    def __init__(self, service):
        self.service = service

    def _check_features(self, report, metrics, dimensions, filters):
        raise NotImplementedError("you need to use a class that inherits from this one")

    def retrieve(self, metrics, start_date, **kwargs):
        if not self.service._service:
            raise NoAuthorisedService("the YouTube service has not been authorised")

        end_date = kwargs.get("end_date", dt.date.today())
        currency = kwargs.get("currency", "USD")
        dimensions = kwargs.get("dimensions", ())
        filters = kwargs.get("filters", {})
        include_historical_data = kwargs.get("include_historical_data", False)
        max_results = kwargs.get("max_results", False)
        sort_by = kwargs.get("sort_by", ())
        sort_descending = kwargs.get("sort_descending", False)
        start_index = kwargs.get("start_index", 1)

        self._check_features(metrics, dimensions, filters)
        if filters:
            filters = tuple(f"{k}=={v}" for k, v in filters.items())

        return YouTubeAnalyticsReport(
            self.service._service.reports()
            .query(
                ids="channel==MINE",
                metrics=",".join(metrics),
                startDate=start_date,
                endDate=end_date,
                currency=currency,
                dimensions=",".join(dimensions),
                filters=";".join(filters),
                includeHistoricalChannelData=include_historical_data,
                maxResults=max_results,
                sort=",".join(sort_by),
                startIndex=start_index,
            )
            .execute()
        )


class Features(metaclass=abc.ABCMeta):
    __slots__ = ("none", "req", "opt", "many")

    def __init__(self, none=False, req=[], opt=[], many=[]):
        self.none = none
        self.req = [set(x) for x in req]
        self.opt = [set(x) for x in opt]
        self.many = [set(x) for x in many]

    @property
    def every(self):
        e = []
        for x in [*[*self.req], *[*self.opt], *[*self.many]]:
            e.extend(x)
        return e

    def verify(self, against):
        raise NotImplementedError()



YOUTUBE_ANALYTICS_API_SERVICE_NAME = "youtubeAnalytics"
YOUTUBE_ANALYTICS_API_VERSION = "v2"
YOUTUBE_ANALYTICS_SCOPES = (
    "yt-analytics.readonly",
    "yt-analytics-monetary.readonly",
)

from .analytics import *
from .service import YouTubeService



class AnalytixError(Exception):
    pass


class NoAuthorisedService(AnalytixError):
    pass


class ServiceAlreadyExists(AnalytixError):
    pass


class IncompleteRequest(AnalytixError):
    pass


class InvalidRequest(AnalytixError):
    pass



__productname__ = "analytix"
__version__ = "0.1.0a1"
__description__ = "analytix is a simple yet powerful API wrapper to make getting analytical information from your favourite services easier than ever."
__url__ = "https://github.com/parafoxia/analytix"
__docs__ = "https://analytix.readthedocs.io/en/latest/"
__author__ = "Ethan Henderson"
__license__ = "BSD-3-Clause"

from .errors import *



import json
import os
import typing as t

from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient import discovery

from analytix import IncompleteRequest, NoAuthorisedService, ServiceAlreadyExists
from analytix.youtube import YOUTUBE_ANALYTICS_API_SERVICE_NAME, YOUTUBE_ANALYTICS_API_VERSION


class YouTubeService:
    __slots__ = ("_service", "_secrets")

    def __init__(self, secrets):
        self._service = None

        if not isinstance(secrets, (os.PathLike, str)):
            self._secrets = secrets
            return

        with open(secrets, "r", encoding="utf-8") as f:
            self._secrets = json.load(f)

    def authorise(self, *scopes, use_console=False):
        if not scopes:
            raise IncompleteRequest("expected 1 or more scopes, got 0")

        if self._service:
            raise ServiceAlreadyExists("an authorised service already exists")

        flow = InstalledAppFlow.from_client_config(
            self._secrets, tuple(f"https://www.googleapis.com/auth/{s}" for s in scopes)
        )
        try:
            credentials = flow.run_local_server(open_browser=True) if not use_console else flow.run_console()
        except OSError:
            print("WARNING: Using console authentication as server authentication failed.")
            credentials = flow.run_console()
        self._service = discovery.build(
            YOUTUBE_ANALYTICS_API_SERVICE_NAME, YOUTUBE_ANALYTICS_API_VERSION, credentials=credentials
        )

    def close(self):
        if not self._service:
            raise NoAuthorisedService("no authorised service currently exists")

        self._service.close()
        self._service = None



from analytix import InvalidRequest
from analytix.youtube.abc import Features

YOUTUBE_ANALYTICS_DEFAULT_METRICS = (
    "views",
    "redViews",
    "comments",
    "likes",
    "dislikes",
    "videosAddedToPlaylists",
    "videosRemovedFromPlaylists",
    "shares",
    "estimatedMinutesWatched",
    "estimatedRedMinutesWatched",
    "averageViewDuration",
    "averageViewPercentage",
    "annotationClickThroughRate",
    "annotationCloseRate",
    "annotationImpressions",
    "annotationClickableImpressions",
    "annotationClosableImpressions",
    "annotationClicks",
    "annotationCloses",
    "cardClickRate",
    "cardTeaserClickRate",
    "cardImpressions",
    "cardTeaserImpressions",
    "cardClicks",
    "cardTeaserClicks",
    "subscribersGained",
    "subscribersLost",
    "estimatedRevenue",
    "estimatedAdRevenue",
    "grossRevenue",
    "estimatedRedPartnerRevenue",
    "monetizedPlaybacks",
    "playbackBasedCpm",
    "adImpressions",
    "cpm",
)
YOUTUBE_ANALYTICS_DEFAULT_PROVINCE_METRICS = (
    "views",
    "redViews",
    "estimatedMinutesWatched",
    "estimatedRedMinutesWatched",
    "averageViewDuration",
    "averageViewPercentage",
    "annotationClickThroughRate",
    "annotationCloseRate",
    "annotationImpressions",
    "annotationClickableImpressions",
    "annotationClosableImpressions",
    "annotationClicks",
    "annotationCloses",
    "cardClickRate",
    "cardTeaserClickRate",
    "cardImpressions",
    "cardTeaserImpressions",
    "cardClicks",
    "cardTeaserClicks",
)


class Dimensions(Features):
    def __init__(self, none=False, req=[], opt=[], many=[]):
        super().__init__(none, req, opt, many)

    def verify(self, against):
        against = set(against)

        if self.none and against:
            raise InvalidRequest(f"expected 0 dimensions, got {len(against)}")

        every = self.every
        for f in against:
            if f not in every:
                raise InvalidRequest(f"unexpected dimension: {f}")

        for req in self.req:
            if len(against & req) != 1:
                raise InvalidRequest(f"expected 1 dimension from {req}, got {len(against)}")

        for opt in self.opt:
            if not 0 <= len(against & opt) <= 1:
                raise InvalidRequest(f"expected 0 or 1 dimensions from {opt}, got {len(against)}")

        for many in self.many:
            if len(against & many) == 0:
                raise InvalidRequest(f"expected at least 1 dimension from {many}, got {len(against)}")


class Filters(Features):
    def __init__(self, none=False, req=[], opt=[], many=[]):
        super().__init__(none, req, opt, many)

    def verify(self, against):
        if not isinstance(against, dict):
            raise InvalidRequest(f"expected dict of features, got {type(against).__name__}")
        against = set(against.keys())

        if self.none and against:
            raise InvalidRequest(f"expected 0 filters, got {len(against)}")

        every = self.every
        for f in against:
            if f not in every:
                raise InvalidRequest(f"unexpected filter: {f}")

        for req in self.req:
            if len(against & req) != 1:
                raise InvalidRequest(f"expected 1 filter from {req}, got {len(against)}")

        for opt in self.opt:
            if not 0 <= len(against & opt) <= 1:
                raise InvalidRequest(f"expected 0 or 1 filters from {opt}, got {len(against)}")

        for many in self.many:
            if len(against & many) == 0:
                raise InvalidRequest(f"expected at least 1 filter from {many}, got {len(against)}")



import datetime as dt
import json
import typing as t

from analytix import InvalidRequest
from analytix.youtube.abc import YouTubeAnalytics
from analytix.youtube.features import (
    YOUTUBE_ANALYTICS_DEFAULT_METRICS,
    YOUTUBE_ANALYTICS_DEFAULT_PROVINCE_METRICS,
    Dimensions,
    Filters,
)


class BasicYouTubeAnalytics(YouTubeAnalytics):
    def __init__(self, service, by_province=False):
        super().__init__(service)
        self.dimensions = Dimensions(none=True)
        self.by_province = by_province

    @property
    def by_province(self):
        return self._by_province

    @by_province.setter
    def by_province(self, value):
        if not isinstance(value, bool):
            raise TypeError(f"expected bool for by_province, got {type(value).__name__}")

        self._by_province = value
        if not value:
            self.metrics = YOUTUBE_ANALYTICS_DEFAULT_METRICS
            self.filters = Filters(opt=[("country", "continent", "subContinent"), ("video", "group")])
        else:
            self.metrics = YOUTUBE_ANALYTICS_DEFAULT_PROVINCE_METRICS
            self.filters = Filters(many=[("province",)], opt=[("video", "group")])

    def _check_features(self, metrics, dimensions, filters):
        if not metrics:
            raise InvalidRequest(f"expected at least 1 metric, got 0")

        for m in metrics:
            if m not in self.metrics:
                raise InvalidRequest(f"unexpected metric: {m}")

        self.dimensions.verify(dimensions)
        self.filters.verify(filters)


class TimeBasedYouTubeAnalytics(YouTubeAnalytics):
    def __init__(self, service, by_province=False):
        super().__init__(service)
        self.dimensions = Dimensions(req=[("day", "month")])
        self.by_province = by_province

    @property
    def by_province(self):
        return self._by_province

    @by_province.setter
    def by_province(self, value):
        if not isinstance(value, bool):
            raise TypeError(f"expected bool for by_province, got {type(value).__name__}")

        self._by_province = value
        if not value:
            self.metrics = YOUTUBE_ANALYTICS_DEFAULT_METRICS
            self.filters = Filters(opt=[("country", "continent", "subContinent"), ("video", "group")])
        else:
            self.metrics = YOUTUBE_ANALYTICS_DEFAULT_PROVINCE_METRICS
            self.filters = Filters(many=[("province",)], opt=[("video", "group")])

    def _check_features(self, metrics, dimensions, filters):
        if not metrics:
            raise InvalidRequest(f"expected at least 1 metric, got 0")

        for m in metrics:
            if m not in self.metrics:
                raise InvalidRequest(f"unexpected metric: {m}")

        self.dimensions.verify(dimensions)
        self.filters.verify(filters)



import setuptools

import analytix

with open("./README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name=analytix.__productname__,
    version=analytix.__version__,
    description=analytix.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=analytix.__url__,
    author=analytix.__author__,
    license=analytix.__license__,
    classifiers=[
        "Development Status :: 3 - Alpha",
        # "Development Status :: 4 - Beta",
        # "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Utilities",
    ],
    project_urls={
        "Documentation": analytix.__docs__,
        "Source": analytix.__url__,
    },
    install_requires=[
        "google-api-python-client>=1.12.0",
        "google-auth-oauthlib>=0.4.2",
        "pandas>=1.2.0",
    ],
    # extras_require={},
    python_requires=">=3.6",
    packages=setuptools.find_packages(exclude=["tests*"]),
)
